import os
from pathlib import Path
from typing import Callable

import pandas as pd

from future_sales.data.utils import transforms


class ETL:
    def __init__(self, input_dir: Path, save_dir: Path) -> None:
        self.input_dir = input_dir
        self.save_dir = save_dir

    def extract(self) -> dict[str, pd.DataFrame]:
        data = {}
        for _, _, filenames in os.walk(self.input_dir):
            for filename in filenames:
                data[filename.split(".")[0]] = pd.read_csv(
                    os.path.join(self.input_dir, filename)
                )

        return data

    def transform(self, data: dict[str, pd.DataFrame]) -> dict[str, pd.DataFrame]:
        def process_df(
            df: pd.DataFrame, handlers: list[Callable[[pd.DataFrame], pd.DataFrame]]
        ) -> pd.DataFrame:
            _df = df.copy()
            for handler in handlers:
                _df = handler(_df)

            return _df

        transformed_data = {
            tbl_name: process_df(
                df=data[tbl_name],
                handlers=transforms[tbl_name] if tbl_name in transforms.keys() else [],
            )
            for tbl_name in data.keys()
        }
        return transformed_data

    def load(self, data: dict[str, pd.DataFrame]) -> None:
        for filename in data.keys():
            data[filename].to_csv(
                os.path.join(self.save_dir, f"{filename}.csv"), index=False
            )



from pathlib import Path
import pickle

import click
import pandas as pd


@click.command()
@click.option(
    "--input-dir",
    default="data/processed/test.csv",
    type=click.Path(exists=True, dir_okay=False),
    show_default=True,
    help="Path to directory with preprocessed test data.",
)
@click.option(
    "--checkpoint-path",
    default="models/model.pkl",
    type=click.Path(exists=True, dir_okay=False),
    show_default=True,
    help="Path to checkpoint.",
)
@click.option(
    "--submission-path",
    default="submissions/submission.csv",
    type=click.Path(dir_okay=False, writable=True),
    show_default=True,
    help="Path to save submission file.",
)
def predict(input_dir: Path, checkpoint_path: Path, submission_path: Path) -> None:
    X = pd.read_csv(input_dir)

    model = pickle.load(open(checkpoint_path, "rb"))

    preds = model.predict(X)

    df = pd.DataFrame(preds, columns=["item_cnt_month"])
    df["ID"] = df.index
    df = df.set_index("ID")
    df.to_csv(submission_path)

    click.echo(f"Submission file is saved to {submission_path}.")



from functools import partial
from typing import Any, Generator, Union

from category_encoders import TargetEncoder
import click
from hyperopt import fmin, hp, tpe
import numpy as np
import numpy.typing as npt
import pandas as pd
from sklearn.metrics import mean_squared_error
from sklearn.model_selection._split import _BaseKFold
from sklearn.utils import indexable
from sklearn.utils.validation import _num_samples
from xgboost import XGBRegressor

from future_sales.data.build_features import build_features

RANDOM_STATE = 42
EARLY_STOPPING_ROUNDS = 10
SPACE = {
    "max_depth": hp.choice("max_depth", np.arange(3, 10, dtype=int)),
    "subsample": hp.uniform("subsample", 0.7, 1),
    "colsample_bytree": hp.uniform("colsample_bytree", 0.7, 1),
    "reg_lambda": hp.uniform("reg_lambda", 0, 1),
}
MAX_EVALS = 20


class GroupTimeSeriesSplit(_BaseKFold):  # type: ignore
    """
    Time Series cross-validator for a variable number of observations within the time unit.
    In the kth split, it returns first k folds as train set and the (k+1)th fold as test set.
    Parameters
    ----------
    n_splits : int, default=5
        Number of splits. Must be at least 2.
    """

    def __init__(self, *, n_splits: int = 5) -> None:
        super().__init__(n_splits, shuffle=False, random_state=None)

    def split(
        self,
        X: npt.NDArray[Union[np.int_, np.float_]],
        groups: npt.NDArray[np.int_],
    ) -> Generator[tuple[npt.NDArray[np.int_], npt.NDArray[np.int_]], None, None]:
        """Generate indices to split data into training and test set.
        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Training data, where `n_samples` is the number of samples
            and `n_features` is the number of features.
        groups : array-like of shape (n_samples,)
            Group labels for the samples used while splitting the dataset into
            train/test set.
        Yields
        ------
        train : ndarray
            The training set indices for that split.
        test : ndarray
            The testing set indices for that split.
        """
        X, groups = indexable(X, groups)
        n_samples = _num_samples(X)
        indices = np.arange(n_samples)
        _, group_counts = np.unique(groups, return_counts=True)
        folds = np.split(indices, np.cumsum(group_counts)[:-1])
        n_folds = _num_samples(folds)
        test_starts = range(1, n_folds, 1)
        for test_start in test_starts:
            yield (
                np.concatenate(folds[:test_start]),
                np.concatenate(folds[test_start : test_start + 1]),
            )


def evaluate(
    data: dict[str, pd.DataFrame]
) -> tuple[float, TargetEncoder, pd.DataFrame]:
    X = data["sales_train"]

    gtscv = GroupTimeSeriesSplit(n_splits=33)
    groups = X["date_block_num"]
    folds = gtscv.split(X, groups=groups)

    cat_cols = [
        "shop_id",
        "item_id",
        "item_category_id",
        "item_first_sale",
        "city_code",
        "shop_type_code",
        "category_group_code",
        "category_subgroup_code",
        "shop_first_month",
        "shop_last_month",
        "year",
        "month",
        "shop_cluster",
        # "item_cluster",
    ]

    cv_scores = []
    preprocessed_dataset = []  # type: ignore
    for train_id, val_id in folds:
        click.echo(f"Current validation set: {np.unique(groups[val_id])}")

        # train set
        if len(preprocessed_dataset) == 0:
            X_train_full = build_features(data, indices=train_id)
        else:
            X_train_full = pd.concat((X_train_full, preprocessed_dataset.pop()))

        X_train, y_train = (
            X_train_full.drop(columns=["item_cnt_month"]),
            X_train_full["item_cnt_month"],
        )
        # y_train = y_train.clip(0, 20)
        click.echo(f"Train dataset shape: {X_train.shape}")

        # val set
        X_val_full = build_features(
            data, indices=np.concatenate((train_id, val_id))
        ).iloc[-len(val_id) :, :]
        preprocessed_dataset.append(X_val_full)
        X_val, y_val = (
            X_val_full.drop(columns=["item_cnt_month"]),
            X_val_full["item_cnt_month"],
        )
        # y_val = y_val.clip(0, 20)
        click.echo(f"Val dataset shape: {X_val.shape}\n")

        te = TargetEncoder(cols=cat_cols)
        te.fit(X_train, y_train)
        X_train = te.transform(X_train)
        X_val = te.transform(X_val)

        model = XGBRegressor(
            eval_metric="rmse",
            early_stopping_rounds=EARLY_STOPPING_ROUNDS,
            random_state=RANDOM_STATE,
            n_jobs=-1,
        )
        model.fit(
            X_train,
            y_train,
            eval_set=[(X_train, y_train), (X_val, y_val)],
            verbose=False,
        )
        cv_scores.append(model.best_score)

    X_train_full = pd.concat((X_train_full, preprocessed_dataset.pop()))

    return np.array(cv_scores).mean(), te, X_train_full


def objective(space: dict[str, pd.DataFrame], data: pd.DataFrame) -> Any:
    X_train, X_val = (
        data[data["date_block_num"] <= 30],
        data[data["date_block_num"] > 30],
    )
    X_train, y_train = (
        X_train.drop(columns=["item_cnt_month"]),
        X_train["item_cnt_month"],
    )
    # y_train = y_train.clip(0, 20)
    X_val, y_val = (
        X_val.drop(columns=["item_cnt_month"]),
        X_val["item_cnt_month"],
    )
    # y_val = y_val.clip(0, 20)

    model = XGBRegressor(
        max_depth=int(space["max_depth"]),
        reg_lambda=space["reg_lambda"],
        subsample=space["subsample"],
        colsample_bytree=space["colsample_bytree"],
        eval_metric="rmse",
        early_stopping_rounds=EARLY_STOPPING_ROUNDS,
        random_state=RANDOM_STATE,
        n_jobs=-1,
    )
    model.fit(
        X_train,
        y_train,
        eval_set=[(X_train, y_train), (X_val, y_val)],
        verbose=False,
    )

    pred = model.predict(X_val)

    rmse = mean_squared_error(y_val, pred, squared=False)
    print(f"RMSE: {rmse}")

    return rmse


def optimize(data: pd.DataFrame) -> Any:
    best = fmin(
        fn=partial(objective, data=data),
        space=SPACE,
        algo=tpe.suggest,
        max_evals=MAX_EVALS,
    )
    return best



import os
from pathlib import Path
import pickle
from warnings import simplefilter

import click
from xgboost import XGBRegressor

from future_sales.data.build_features import build_features
from future_sales.data.etl import ETL
from future_sales.utils import evaluate, optimize, RANDOM_STATE

simplefilter(action="ignore", category=FutureWarning)


@click.command()
@click.option(
    "--input-dir",
    default="data/raw",
    type=click.Path(exists=True, dir_okay=True),
    show_default=True,
    help="Path to directory with raw data.",
)
@click.option(
    "--interim-dir",
    default="data/intermediate",
    type=click.Path(exists=True, dir_okay=True),
    show_default=True,
    help="Path to directory to save intermediate data that has been transformed.",
)
@click.option(
    "--save-dir",
    default="data/processed",
    type=click.Path(exists=True, dir_okay=True),
    show_default=True,
    help="Path to directory to save final data sets for modeling.",
)
@click.option(
    "--checkpoint-path",
    default="models/model.pkl",
    type=click.Path(dir_okay=False, writable=True),
    show_default=True,
    help="Path to save trained model.",
)
def train(
    input_dir: Path,
    interim_dir: Path,
    save_dir: Path,
    checkpoint_path: Path,
) -> None:
    # extract data
    if len(os.listdir(interim_dir)) == 0:
        etl = ETL(input_dir=input_dir, save_dir=interim_dir)
        click.echo("Extracting data...")
        data = etl.extract()
        click.echo("Transforming data...")
        data = etl.transform(data)
        click.echo("Saving transformed data...\n")
        etl.load(data)
    else:
        click.echo("Extracting data...")
        etl = ETL(input_dir=interim_dir, save_dir=interim_dir)
        data = etl.extract()

    # evaluate
    click.echo("Evaluating...")
    rmse, target_encoder, X_train = evaluate(data)
    click.echo(f"Validation RMSE: {rmse}")

    X_train.to_csv(os.path.join(save_dir, "train.csv"), index=False)

    # X_train = pd.read_csv(os.path.join(save_dir, "train.csv"))
    # X_test = pd.read_csv(os.path.join(save_dir, "test.csv"))

    # tune hyperparameters
    click.echo("Hyperparameters tuning...")
    best_params = optimize(X_train)
    print(f"Best hyperparameters: {best_params}")

    # train
    click.echo("Training...")
    X_train, y_train = (
        X_train.drop(columns=["item_cnt_month"]),
        X_train["item_cnt_month"],
    )
    # y_train = y_train.clip(0, 20)
    click.echo(f"Train dataset shape: {X_train.shape}")

    X_test = build_features(data, indices=list(range(X_train.shape[0])), is_test=True)
    X_test = target_encoder.transform(X_test)
    X_test.to_csv(os.path.join(save_dir, "test.csv"), index=False)

    model = XGBRegressor(
        **best_params,
        eval_metric="rmse",
        early_stopping_rounds=None,
        random_state=RANDOM_STATE,
        n_jobs=-1,
    )
    model.fit(X_train, y_train)
    pickle.dump(model, open(checkpoint_path, "wb"))
    click.echo(f"Model is saved to {checkpoint_path}")
